/*
 * libleaftag-gtk/tag-dialog.c - A GtkDialog for tagging URIs
 *
 * Copyright (C) 2006 David Trowbridge
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 */
#include <libleaftag-gtk/tag-dialog.h>
#include <libleaftag-gtk/tag-list.h>
#include <libleaftag/leaftag.h>
#include <gtk/gtk.h>
#include <glib/gi18n.h>
#include <string.h>

struct _LeaftagGtkTagDialogPriv
{
	guint        n_uris;
	GList       *uris;

	GtkTooltips *tooltips;

	GHashTable  *bundle_widgets;
	GHashTable  *bundle_expanders;

	GtkWidget   *vbox;
	GtkWidget   *tag_entry;
	GtkWidget   *soup_toggle;
	GtkWidget   *items_label;
	GtkWidget   *files_label;
	GtkWidget   *soup_hbox;
	GtkWidget   *soup_vbox;
	GtkWidget   *scrolled_window;

	GList       *all_tags;
	GList       *set_tags;
	GList       *inconsistent_tags;
	GList       *new_tags;

	guint        changed_handler;
	guint        insert_handler;

	gint         default_height;
};

static void       leaftag_gtk_tag_dialog_class_init (LeaftagGtkTagDialogClass *klass);
static void       leaftag_gtk_tag_dialog_init       (LeaftagGtkTagDialog      *dialog);
static void       leaftag_gtk_tag_dialog_finalize   (GObject                  *obj);
static void       leaftag_gtk_tag_dialog_response   (GtkDialog                *dialog,
                                                     gint                      response_id);
static void       create_soup                       (LeaftagGtkTagDialog      *dialog);
static GtkWidget *add_bundle                        (LeaftagGtkTagDialog      *dialog,
                                                     const gchar              *bundle,
                                                     gboolean                  synthetic);
static void       create_user_bundles               (LeaftagGtkTagDialog      *dialog);
static void       soup_toggled                      (GtkToggleButton          *button,
                                                     LeaftagGtkTagDialog      *dialog);
static void       tag_entry_changed                 (GtkEntry                 *entry,
                                                     LeaftagGtkTagDialog      *dialog);
static void       tag_entry_insert                  (GtkEntry                 *entry,
                                                     gchar                    *new_text,
                                                     gint                      new_text_length,
                                                     gint                     *position,
                                                     LeaftagGtkTagDialog      *dialog);
static void       tag_button_activated              (LeaftagGtkTagList        *bundle,
                                                     gchar                    *tag,
                                                     LeaftagGtkTagDialog      *dialog);
static void       tag_button_deactivated            (LeaftagGtkTagList        *bundle,
                                                     gchar                    *tag,
                                                     LeaftagGtkTagDialog      *dialog);
static void       remove_from_all_bundles           (LeaftagGtkTagDialog      *dialog,
                                                     gchar                    *tag);
static void       set_in_all_bundles                (LeaftagGtkTagDialog      *dialog,
                                                     gchar                    *tag,
                                                     gboolean                  active);
static void       set_inconsistent                  (LeaftagGtkTagDialog      *dialog,
                                                     gchar                    *tag);
static void       add_to_all_tags_bundle            (LeaftagGtkTagDialog      *dialog,
                                                     gchar                    *tag);
static void       freeze_all_bundles                (LeaftagGtkTagDialog      *dialog);
static void       thaw_all_bundles                  (LeaftagGtkTagDialog      *dialog);
static void       set_existing_tags                 (LeaftagGtkTagDialog      *dialog);

static GtkDialog *parent_class = NULL;

G_DEFINE_TYPE (LeaftagGtkTagDialog, leaftag_gtk_tag_dialog, GTK_TYPE_DIALOG);

static void
leaftag_gtk_tag_dialog_class_init (LeaftagGtkTagDialogClass *klass)
{
	GObjectClass   *object_class;
	GtkDialogClass *dialog_class;

	parent_class = g_type_class_peek_parent (klass);

	object_class = G_OBJECT_CLASS(klass);
	object_class->finalize = leaftag_gtk_tag_dialog_finalize;

	dialog_class = GTK_DIALOG_CLASS(klass);
	dialog_class->response = leaftag_gtk_tag_dialog_response;
}

static void
leaftag_gtk_tag_dialog_init (LeaftagGtkTagDialog *dialog)
{
	GtkWidget      *entry_hbox;
	GtkWidget      *title_hbox;
	GtkWidget      *soup_toggle_icon;
	GtkWidget      *tags_label;
	GtkWidget      *junk_label;
	GtkSizeGroup   *size_group;
	GList          *tags;
	GList          *i;
	GtkRequisition  size_request;
	GdkGeometry     geometry;

	dialog->priv = g_new0 (LeaftagGtkTagDialogPriv, 1);

	gtk_dialog_add_buttons (GTK_DIALOG (dialog),
			        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				_("Tag"),         GTK_RESPONSE_OK,
				NULL);
	gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);
	gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
	gtk_window_set_default_size (GTK_WINDOW (dialog), 450, -1);

	dialog->priv->vbox = gtk_vbox_new (FALSE, 6);
	gtk_widget_show (dialog->priv->vbox);
	gtk_container_add (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), dialog->priv->vbox);
	gtk_container_set_border_width (GTK_CONTAINER (dialog->priv->vbox), 6);

	size_group = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);

	/* Create the files label row */
	title_hbox = gtk_hbox_new (FALSE, 6);
	dialog->priv->items_label = gtk_label_new (_("Items:"));
	dialog->priv->files_label = gtk_label_new ("");
	gtk_widget_show (title_hbox);
	gtk_widget_show (dialog->priv->items_label);
	gtk_widget_show (dialog->priv->files_label);
	gtk_size_group_add_widget (size_group, dialog->priv->items_label);
	gtk_misc_set_alignment (GTK_MISC (dialog->priv->items_label), 0.0, 0.5);
	gtk_misc_set_alignment (GTK_MISC (dialog->priv->files_label), 0.0, 0.5);
	gtk_box_pack_start (GTK_BOX (title_hbox), dialog->priv->items_label, FALSE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX (title_hbox), dialog->priv->files_label, TRUE, TRUE, 0);

	gtk_box_pack_start (GTK_BOX (dialog->priv->vbox), title_hbox, FALSE, TRUE, 0);

	/* Create the tags entry row */
	entry_hbox = gtk_hbox_new (FALSE, 6);
	tags_label = gtk_label_new (_("Tags:"));
	dialog->priv->tag_entry = gtk_entry_new ();
	dialog->priv->soup_toggle = gtk_toggle_button_new ();
	soup_toggle_icon = gtk_image_new_from_stock (GTK_STOCK_INDEX, GTK_ICON_SIZE_BUTTON);
	gtk_widget_show (entry_hbox);
	gtk_widget_show (tags_label);
	gtk_widget_show (dialog->priv->tag_entry);
	gtk_widget_show (dialog->priv->soup_toggle);
	gtk_widget_show (soup_toggle_icon);
	gtk_size_group_add_widget (size_group, tags_label);
	gtk_misc_set_alignment (GTK_MISC (tags_label), 0.0, 0.5);
	gtk_container_add (GTK_CONTAINER (dialog->priv->soup_toggle), soup_toggle_icon);
	gtk_box_pack_start (GTK_BOX (entry_hbox), tags_label, FALSE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX (entry_hbox), dialog->priv->tag_entry, TRUE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX (entry_hbox), dialog->priv->soup_toggle, FALSE, TRUE, 0);
	dialog->priv->changed_handler = g_signal_connect (G_OBJECT (dialog->priv->tag_entry), "changed", G_CALLBACK (tag_entry_changed), dialog);
	dialog->priv->insert_handler = g_signal_connect (G_OBJECT (dialog->priv->tag_entry), "insert-text", G_CALLBACK (tag_entry_insert), dialog);
	g_signal_connect (G_OBJECT (dialog->priv->soup_toggle), "toggled", G_CALLBACK (soup_toggled), dialog);

	gtk_box_pack_start (GTK_BOX (dialog->priv->vbox), entry_hbox, FALSE, TRUE, 0);

	/* Create the button soup row.  We don't pack it yet. */
	dialog->priv->soup_hbox = gtk_hbox_new (FALSE, 6);
	g_object_ref (G_OBJECT (dialog->priv->soup_hbox));
	junk_label = gtk_label_new ("");
	dialog->priv->scrolled_window = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_show (dialog->priv->soup_hbox);
	gtk_widget_show (junk_label);
	gtk_widget_show (dialog->priv->scrolled_window);
	gtk_size_group_add_widget (size_group, junk_label);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (dialog->priv->scrolled_window), GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	gtk_box_pack_start (GTK_BOX (dialog->priv->soup_hbox), junk_label, FALSE, TRUE, 0);
	gtk_box_pack_start (GTK_BOX (dialog->priv->soup_hbox), dialog->priv->scrolled_window, TRUE, TRUE, 0);

	/* Set up tooltips for everything */
	dialog->priv->tooltips = gtk_tooltips_new ();
	gtk_tooltips_set_tip (dialog->priv->tooltips, dialog->priv->tag_entry,
	                      _("Enter tags, separated by spaces"),
	                      _("Type tag names here, separated by spaces"));
	gtk_tooltips_set_tip (dialog->priv->tooltips, dialog->priv->soup_toggle,
	                      _("View all tags"),
	                      _("Click to show all existing tags"));

	/* look up all tags */
	tags = lt_get_all_tags ();
	for (i = tags; i; i = g_list_next (i)) {
		const char *tag;

		tag = lt_tag_get_name (LT_TAG (i->data));
		dialog->priv->all_tags = g_list_prepend (dialog->priv->all_tags, g_strdup (tag));
		g_object_unref (G_OBJECT (i->data));
	}
	g_list_free (tags);

	/* hide the soup toggle if there are no existing tags */
	if (dialog->priv->all_tags == NULL)
		gtk_widget_hide (dialog->priv->soup_toggle);

	/* Set default height */
	gtk_widget_size_request (GTK_WIDGET (dialog), &size_request);
	dialog->priv->default_height = size_request.height;
	geometry.max_width = 10000;
	geometry.max_height = dialog->priv->default_height;
	gtk_window_set_geometry_hints (GTK_WINDOW (dialog), GTK_WIDGET (dialog), &geometry, GDK_HINT_MAX_SIZE);
}

static void
leaftag_gtk_tag_dialog_finalize (GObject *obj)
{
	LeaftagGtkTagDialog *dialog;

	g_return_if_fail (obj != NULL);
	g_return_if_fail (LEAFTAG_GTK_IS_TAG_DIALOG (obj));

	dialog = LEAFTAG_GTK_TAG_DIALOG (obj);

	g_list_foreach (dialog->priv->uris,              (GFunc) g_free, NULL);
	g_list_foreach (dialog->priv->all_tags,          (GFunc) g_free, NULL);
	g_list_foreach (dialog->priv->set_tags,          (GFunc) g_free, NULL);
	g_list_foreach (dialog->priv->inconsistent_tags, (GFunc) g_free, NULL);
	g_list_foreach (dialog->priv->new_tags,          (GFunc) g_free, NULL);

	g_list_free (dialog->priv->uris);
	g_list_free (dialog->priv->all_tags);
	g_list_free (dialog->priv->set_tags);
	g_list_free (dialog->priv->inconsistent_tags);
	g_list_free (dialog->priv->new_tags);

	g_hash_table_destroy (dialog->priv->bundle_widgets);
	g_hash_table_destroy (dialog->priv->bundle_expanders);

	g_object_unref (GTK_TOOLTIPS (dialog->priv->tooltips));

	g_free (dialog->priv);

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (obj);
}

static void
leaftag_gtk_tag_dialog_response (GtkDialog *dialog, gint response_id)
{
	LeaftagGtkTagDialog *tag_dialog;

	g_return_if_fail (dialog != NULL);
	g_return_if_fail (LEAFTAG_GTK_IS_TAG_DIALOG (dialog));

	tag_dialog = LEAFTAG_GTK_TAG_DIALOG (dialog);

	if (response_id == GTK_RESPONSE_OK) {
		const gchar  *text;
		gchar       **words;
		gint          i;
		GList        *li, *sources, *tags;

		/* Unset old tags */
		sources = lt_source_lookup_many (tag_dialog->priv->uris);
		for (li = sources; li; li = g_list_next (li)) {
			tags = lt_source_get_tags (LT_SOURCE (li->data));
			lt_source_untag (LT_SOURCE (li->data), tags);
			g_list_foreach (tags, (GFunc) g_object_unref, NULL);
			g_list_free (tags);
		}

		/* Set new tags */
		text = gtk_entry_get_text (GTK_ENTRY (tag_dialog->priv->tag_entry));
		words = g_strsplit (text, ",", 0);

		tags = NULL;
		for (i = 0; words[i]; i++) {
			if (strlen (words[i]) == 0)
				return;
			tags = g_list_prepend (tags, words[i]);
		}
		for (li = sources; li; li = g_list_next (li)) {
			lt_source_tag_with_names (LT_SOURCE (li->data), tags);
		}
		g_strfreev (words);
		g_list_foreach (sources, (GFunc) g_object_unref, NULL);
		g_list_free (sources);
		g_list_free (tags);
	}

	if (GTK_DIALOG_CLASS (parent_class)->response)
		GTK_DIALOG_CLASS (parent_class)->response (dialog, response_id);
}

GtkWidget *
leaftag_gtk_tag_dialog_new (void)
{
	return GTK_WIDGET (g_object_new (LEAFTAG_GTK_TYPE_TAG_DIALOG, 0));
}

void
leaftag_gtk_tag_dialog_set_uris (LeaftagGtkTagDialog *dialog, GList *uris)
{
	GList *is;
	gchar  *text, *tmp;
	GList  *i;
	int     length;

	g_return_if_fail (uris != NULL);

	for (i = uris; i; i = g_list_next (i)) {
		dialog->priv->uris = g_list_prepend (dialog->priv->uris, g_strdup (i->data));
	}
	dialog->priv->uris = g_list_sort (dialog->priv->uris, (GCompareFunc) strcmp);

	set_existing_tags (dialog);

	text = NULL;
	for (is = dialog->priv->set_tags; is; is = g_list_next (is)) {
		if (text) {
			tmp = g_strdup_printf ("%s,%s", text, (gchar *) (is->data));
			g_free (text);
			text = tmp;
		} else {
			text = g_strdup ((gchar *) (is->data));
		}
	}
	g_signal_handler_block (G_OBJECT (dialog->priv->tag_entry), dialog->priv->changed_handler);
	if (text) {
		gtk_entry_set_text (GTK_ENTRY (dialog->priv->tag_entry), text);
		g_free (text);
	}
	g_signal_handler_unblock (G_OBJECT (dialog->priv->tag_entry), dialog->priv->changed_handler);
	gtk_widget_grab_focus (dialog->priv->tag_entry);
	gtk_editable_set_position (GTK_EDITABLE (dialog->priv->tag_entry), -1);

	/* FIXME: it would be nice if this showed something better than URIs for files */
	length = g_list_length (uris);
	if (length < 3) {
		text = NULL;
		for (i = uris; i; i = g_list_next (i)) {
			if (text) {
				tmp = g_strdup_printf ("%s, %s", text, (gchar *) (i->data));
				g_free (text);
				text = tmp;
			} else {
				text = g_strdup ((gchar *) (i->data));
			}
		}
	} else {
		text = g_strdup_printf (_("%d items"), length);
	}
	gtk_label_set_text (GTK_LABEL (dialog->priv->files_label), text);
	g_free (text);
}

static GtkWidget *
add_bundle (LeaftagGtkTagDialog *dialog, const gchar *bundle, gboolean synthetic)
{
	GtkWidget *expander;
	GtkWidget *hbox;
	GtkWidget *junk_label;
	GtkWidget *soup;
	gchar     *label;
	gchar     *key;

	label = g_strdup_printf ("<b>%s</b>", bundle);
	expander = gtk_expander_new (label);
	soup = leaftag_gtk_tag_list_new ();
	g_free (label);
	hbox = gtk_hbox_new (FALSE, 12);
	junk_label = gtk_label_new ("");
	gtk_widget_show (expander);
	gtk_widget_show (hbox);
	gtk_widget_show (junk_label);
	gtk_widget_show (soup);
	gtk_expander_set_use_markup (GTK_EXPANDER (expander), TRUE);
	gtk_expander_set_expanded (GTK_EXPANDER (expander), TRUE);
	gtk_container_add (GTK_CONTAINER (expander), hbox);
	gtk_box_pack_start (GTK_BOX (hbox), junk_label, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (hbox), soup, TRUE, TRUE, 0);

	gtk_box_pack_start (GTK_BOX (dialog->priv->soup_vbox), expander, FALSE, TRUE, 0);

	key = (synthetic) ? g_strdup_printf ("#%s", bundle) : g_strdup (bundle);
	g_hash_table_insert (dialog->priv->bundle_expanders, key, expander);

	key = (synthetic) ? g_strdup_printf ("#%s", bundle) : g_strdup (bundle);
	g_hash_table_insert (dialog->priv->bundle_widgets, key, soup);

	g_signal_connect (G_OBJECT (soup), "tag-activated",   G_CALLBACK (tag_button_activated),   dialog);
	g_signal_connect (G_OBJECT (soup), "tag-deactivated", G_CALLBACK (tag_button_deactivated), dialog);

	return soup;
}

static void
create_user_bundles (LeaftagGtkTagDialog *dialog)
{
	GList *bundles, *i;

	bundles = lt_get_sources_with_schema ("bundle");

	for (i = bundles; i; i = g_list_next (i)) {
		const char *uri;
		const char *bundle_name;
		GtkWidget  *soup;
		GList      *tags;
		GList      *tag;

		uri = lt_source_get_uri (LT_SOURCE (i->data));
		bundle_name = uri + 9; /* Chop out bundle:// -- this is a little gross, but oh well */

		soup = add_bundle (dialog, bundle_name, FALSE);

		tags = lt_source_get_tags (LT_SOURCE (i->data));
		for (tag = tags; tag; tag = g_list_next (tag)) {
			leaftag_gtk_tag_list_add_tag (LEAFTAG_GTK_TAG_LIST (soup),
			                              lt_tag_get_name (LT_TAG (tag->data)));
			g_object_unref (G_OBJECT (tag->data));
		}
		g_list_free (tags);

		g_object_unref (G_OBJECT (i->data));
	}
	g_list_free (bundles);
}

static void
soup_toggled (GtkToggleButton *button, LeaftagGtkTagDialog *dialog)
{
	gint        position;
	GdkGeometry geometry;

	geometry.max_width = 10000;

	if (gtk_toggle_button_get_active (button)) {
		if (dialog->priv->soup_vbox == NULL)
			create_soup (dialog);
		geometry.max_height = 10000;
		gtk_box_pack_start (GTK_BOX (dialog->priv->vbox), dialog->priv->soup_hbox, TRUE, TRUE, 0);
		gtk_widget_show (dialog->priv->soup_hbox);
	} else {
		gtk_widget_hide (dialog->priv->soup_hbox);
		geometry.max_height = dialog->priv->default_height;
		gtk_container_remove (GTK_CONTAINER (dialog->priv->vbox), dialog->priv->soup_hbox);
	}

	gtk_window_set_geometry_hints (GTK_WINDOW (dialog), GTK_WIDGET (dialog), &geometry, GDK_HINT_MAX_SIZE);

	/* Give focus back to the entry.  When the entry grabs focus, it
	 * selects all of its contents, which is bad, so work around that */
	position = gtk_editable_get_position (GTK_EDITABLE (dialog->priv->tag_entry));
	gtk_widget_grab_focus (dialog->priv->tag_entry);
	gtk_editable_set_position (GTK_EDITABLE (dialog->priv->tag_entry), position);
}

static void
create_soup (LeaftagGtkTagDialog *dialog)
{
	GtkWidget *bundle;
	GList     *i;
	GList     *is;

	dialog->priv->soup_vbox = gtk_vbox_new (FALSE, 6);
	gtk_widget_show (dialog->priv->soup_vbox);
	gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (dialog->priv->scrolled_window), dialog->priv->soup_vbox);

	/* Create bundles */
	dialog->priv->bundle_widgets   = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
	dialog->priv->bundle_expanders = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);

	/* Add "all tags" synthetic bundle" */
	bundle = add_bundle (dialog, "All tags", TRUE);
	for (i = dialog->priv->all_tags; i; i = g_list_next (i)) {
		leaftag_gtk_tag_list_add_tag (LEAFTAG_GTK_TAG_LIST (bundle), (gchar *) (i->data));
	}

	/* Add all user-created bundles */
	create_user_bundles (dialog);

	/* Add new tags */
	for (is = dialog->priv->new_tags; is; is = g_list_next (is))
		add_to_all_tags_bundle (dialog, (gchar *) (is->data));

	/* Set buttons properly */
	for (is = dialog->priv->set_tags; is; is = g_list_next (is))
		set_in_all_bundles (dialog, (gchar *) (is->data), TRUE);
	for (is = dialog->priv->inconsistent_tags; is; is = g_list_next (is))
		set_inconsistent (dialog, (gchar *) (is->data));
}

static void
tag_button_activated (LeaftagGtkTagList *bundle, gchar *tag, LeaftagGtkTagDialog *dialog)
{
	const gchar *text;

	g_signal_handler_block (G_OBJECT (dialog->priv->tag_entry), dialog->priv->changed_handler);

	text = gtk_entry_get_text (GTK_ENTRY (dialog->priv->tag_entry));
	if (strlen (text) == 0) {
		gtk_entry_set_text (GTK_ENTRY (dialog->priv->tag_entry), tag);
	} else {
		gchar *newtext;
		newtext = g_strdup_printf ("%s,%s", text, tag);
		gtk_entry_set_text (GTK_ENTRY (dialog->priv->tag_entry), newtext);
		g_free (newtext);
	}
	gtk_widget_grab_focus (dialog->priv->tag_entry);
	gtk_editable_select_region (GTK_EDITABLE (dialog->priv->tag_entry), 0, 0);
	gtk_editable_set_position (GTK_EDITABLE (dialog->priv->tag_entry), -1);

	set_in_all_bundles (dialog, tag, TRUE);

	g_signal_handler_unblock (G_OBJECT (dialog->priv->tag_entry), dialog->priv->changed_handler);
}

static void
tag_button_deactivated (LeaftagGtkTagList *bundle, gchar *tag, LeaftagGtkTagDialog *dialog)
{
	const gchar  *text;
	gchar       **words;
	gchar        *newtext, *tmp;
	gint          i;

	g_signal_handler_block (G_OBJECT (dialog->priv->tag_entry), dialog->priv->changed_handler);

	text = gtk_entry_get_text (GTK_ENTRY (dialog->priv->tag_entry));
	words = g_strsplit (text, ",", 0);

	newtext = NULL;
	for (i = 0; words[i]; i++) {
		if (strlen (words[i]) == 0)
			continue;
		if (strcmp (words[i], tag) == 0)
			continue;
		if (newtext) {
			tmp = g_strdup_printf ("%s,%s", newtext, words[i]);
			g_free (newtext);
			newtext = tmp;
		} else {
			newtext = g_strdup (words[i]);
		}
	}
	if (newtext) {
		gtk_entry_set_text (GTK_ENTRY (dialog->priv->tag_entry), newtext);
		g_free (newtext);
	} else {
		gtk_entry_set_text (GTK_ENTRY (dialog->priv->tag_entry), "");
	}
	gtk_widget_grab_focus (dialog->priv->tag_entry);
	gtk_editable_set_position (GTK_EDITABLE (dialog->priv->tag_entry), -1);

	g_strfreev (words);

	if (g_list_find_custom (dialog->priv->new_tags, tag, (GCompareFunc) strcmp))
		remove_from_all_bundles (dialog, tag);

	set_in_all_bundles (dialog, tag, FALSE);

	g_signal_handler_unblock (G_OBJECT (dialog->priv->tag_entry), dialog->priv->changed_handler);
}

static void
tag_entry_changed (GtkEntry *entry, LeaftagGtkTagDialog *dialog)
{
	const gchar  *text;
	gchar       **words;
	int           i;
	GList        *li;
	gint          position;
	gint          length;

	freeze_all_bundles (dialog);

	/* Start by removing all old "new" tags.  These will get readded later */
	for (li = dialog->priv->new_tags; li; li = g_list_next (li)) {
		remove_from_all_bundles (dialog, li->data);
		g_free (li->data);
	}
	g_list_free (dialog->priv->new_tags);
	dialog->priv->new_tags = NULL;

	/* Now deactivate all tags.  These get reactivated later */
	for (li = dialog->priv->all_tags; li; li = g_list_next (li)) {
		set_in_all_bundles (dialog, li->data, FALSE);
	}

	position = gtk_editable_get_position (GTK_EDITABLE (entry));

	text = gtk_entry_get_text (entry);
	words = g_strsplit (text, ",", 0);

	length = 0;
	for (i = 0; words[i]; i++) {
		gint wordlen;

		wordlen = g_utf8_strlen (words[i], -1);
		if (wordlen == 0) {
			length++;
			continue;
		}

		if (g_list_find_custom (dialog->priv->all_tags, words[i], (GCompareFunc) strcmp)) {
			set_in_all_bundles (dialog, words[i], TRUE);
		} else {
			gchar *dupe;

			dupe = g_strdup (words[i]);
			dialog->priv->new_tags = g_list_prepend (dialog->priv->new_tags, dupe);
			add_to_all_tags_bundle (dialog, words[i]);
		}

		length += (wordlen + 1);
	}

	thaw_all_bundles (dialog);

	g_strfreev (words);
}

static void
tag_entry_insert (GtkEntry *entry, gchar *new_text, gint new_text_length, gint *position, LeaftagGtkTagDialog *dialog)
{
	gchar *clean_text;
	gint   i, j;

	clean_text = g_strndup (new_text, new_text_length);
	for (i = 0, j = 0; i < new_text_length; i++) {
		gchar c = new_text[i];
		if (c == '\t' || c == '\n' || c == '\r' || c == '(' ||
		    c == ')'  || c == '['  || c == ']'  || c == '<' ||
		    c == '>'  || c == '+'  || c == '\'' || c == '"')
			continue;
		clean_text[j] = c;
		j++;
	}
	clean_text[j] = '\0';

	/* Call the default handler directly and stop the emission */
	g_signal_stop_emission_by_name (G_OBJECT (entry), "insert-text");
	if (clean_text[0] != '\0') {
		GtkEditableClass *klass;
		klass = GTK_EDITABLE_GET_CLASS (entry);
		klass->insert_text (GTK_EDITABLE (entry), clean_text, j, position);
	}

	g_free (clean_text);
}

static void
remove_from_all_bundles_iterate (gchar *name, GtkWidget *bundle, gchar *tag)
{
	leaftag_gtk_tag_list_remove_tag (LEAFTAG_GTK_TAG_LIST (bundle), tag);
}

static void
remove_from_all_bundles (LeaftagGtkTagDialog *dialog, gchar *tag)
{
	if (dialog->priv->bundle_widgets == NULL)
		return;

	g_hash_table_foreach (dialog->priv->bundle_widgets, (GHFunc) remove_from_all_bundles_iterate, tag);
}

static void
set_active_in_all_bundles_iterate (gchar *name, GtkWidget *bundle, gchar *tag) {
	leaftag_gtk_tag_list_set_active (LEAFTAG_GTK_TAG_LIST (bundle), tag);
}

static void
set_inactive_in_all_bundles_iterate (gchar *name, GtkWidget *bundle, gchar *tag) {
	leaftag_gtk_tag_list_set_inactive (LEAFTAG_GTK_TAG_LIST (bundle), tag);
}

static void
set_in_all_bundles (LeaftagGtkTagDialog *dialog, gchar *tag, gboolean active)
{
	if (dialog->priv->bundle_widgets == NULL)
		return;

	if (active)
		g_hash_table_foreach (dialog->priv->bundle_widgets, (GHFunc) set_active_in_all_bundles_iterate, tag);
	else
		g_hash_table_foreach (dialog->priv->bundle_widgets, (GHFunc) set_inactive_in_all_bundles_iterate, tag);
}

static void
set_inconsistent_iterate (gchar *name, GtkWidget *bundle, gchar *tag) {
	leaftag_gtk_tag_list_set_inconsistent (LEAFTAG_GTK_TAG_LIST (bundle), tag);
}

static void
set_inconsistent (LeaftagGtkTagDialog *dialog, gchar *tag)
{
	g_hash_table_foreach (dialog->priv->bundle_widgets, (GHFunc) set_inconsistent_iterate, tag);
}

static void
add_to_all_tags_bundle (LeaftagGtkTagDialog *dialog, gchar *tag)
{
	GtkWidget *bundle;

	if (dialog->priv->bundle_widgets == NULL)
		return;

	bundle = g_hash_table_lookup (dialog->priv->bundle_widgets, "#All tags");

	leaftag_gtk_tag_list_add_tag (LEAFTAG_GTK_TAG_LIST (bundle), tag);
	leaftag_gtk_tag_list_set_active (LEAFTAG_GTK_TAG_LIST (bundle), tag);
}

static void
freeze_all_bundles_iterate (gchar *name, GtkWidget *bundle, gpointer data)
{
	leaftag_gtk_tag_list_freeze (LEAFTAG_GTK_TAG_LIST (bundle));
}

static void
freeze_all_bundles (LeaftagGtkTagDialog *dialog)
{
	if (dialog->priv->bundle_widgets == NULL)
		return;

	g_hash_table_foreach (dialog->priv->bundle_widgets, (GHFunc) freeze_all_bundles_iterate, NULL);
}

static void
thaw_all_bundles_iterate (gchar *name, GtkWidget *bundle, gpointer data)
{
	leaftag_gtk_tag_list_thaw (LEAFTAG_GTK_TAG_LIST (bundle));
}

static void
thaw_all_bundles (LeaftagGtkTagDialog *dialog)
{
	if (dialog->priv->bundle_widgets == NULL)
		return;

	g_hash_table_foreach (dialog->priv->bundle_widgets, (GHFunc) thaw_all_bundles_iterate, NULL);
}

static void
set_existing_iterate (gchar *tag, gpointer countp, LeaftagGtkTagDialog *dialog)
{
	guint count;

	count = GPOINTER_TO_UINT (countp);

	if (count == dialog->priv->n_uris) {
		dialog->priv->set_tags = g_list_append (dialog->priv->set_tags, g_strdup (tag));
	} else if (count != 0) {
		dialog->priv->inconsistent_tags = g_list_append (dialog->priv->set_tags, g_strdup (tag));
	}
}

static void
set_existing_tags (LeaftagGtkTagDialog *dialog)
{
	GHashTable *tag_counts;
	GList      *i, *j;
	GList      *sources;

	tag_counts = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);

	sources = lt_source_lookup_many (dialog->priv->uris);
	for (i = sources; i; i = g_list_next (i)) {
		GList *tags;

		tags = lt_source_get_tags (LT_SOURCE (i->data));
		for (j = tags; j; j = g_list_next (j)) {
			const gchar *name;
			guint  count;

			name = lt_tag_get_name (LT_TAG (j->data));
			count = GPOINTER_TO_UINT (g_hash_table_lookup (tag_counts, name));
			count += 1;
			g_hash_table_insert (tag_counts, g_strdup (name), GUINT_TO_POINTER (count));

			g_object_unref (G_OBJECT (j->data));
		}
		g_list_free (tags);
	}

	dialog->priv->n_uris = g_list_length (dialog->priv->uris);
	g_hash_table_foreach (tag_counts, (GHFunc) set_existing_iterate, dialog);

	g_hash_table_destroy (tag_counts);
}
