/*
 * libleaftag-gtk/tag-list.h - A widget which keeps a list of tags, shown as
 *                             GtkToggleButtons, with reflowing
 *                             height-for-width sizing.
 *
 * Copyright (C) 2006 David Trowbridge
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 */
#include <libleaftag-gtk/tag-list.h>
#include <gtk/gtk.h>
#include <string.h>

struct _LeaftagGtkTagListPriv
{
	GSList     *tags;
	GSList     *rows;
	GHashTable *buttons;
	GHashTable *button_widths;
	GHashTable *button_handlers;
	gint        width;

	gboolean    frozen;
};

static void leaftag_gtk_tag_list_class_init    (LeaftagGtkTagListClass *klass);
static void leaftag_gtk_tag_list_init          (LeaftagGtkTagList      *list);
static void leaftag_gtk_tag_list_finalize      (GObject                *obj);
static void leaftag_gtk_tag_list_size_allocate (GtkWidget              *widget,
                                                GtkAllocation          *allocation);
static void reflow                             (LeaftagGtkTagList      *list);
static void button_toggled                     (GtkToggleButton        *button,
                                                LeaftagGtkTagList      *list);

enum {
	TAG_ACTIVATED,
	TAG_DEACTIVATED,
	LAST_SIGNAL
};
static guint signals[LAST_SIGNAL] = {0};

static GtkVBoxClass *parent_class = NULL;

G_DEFINE_TYPE (LeaftagGtkTagList, leaftag_gtk_tag_list, GTK_TYPE_VBOX);

static void
leaftag_gtk_tag_list_class_init (LeaftagGtkTagListClass *klass)
{
	GObjectClass   *object_class;
	GtkWidgetClass *widget_class;

	parent_class = g_type_class_peek_parent (klass);

	object_class = G_OBJECT_CLASS(klass);
	object_class->finalize = leaftag_gtk_tag_list_finalize;

	widget_class = GTK_WIDGET_CLASS(klass);
	widget_class->size_allocate = leaftag_gtk_tag_list_size_allocate;

	signals[TAG_ACTIVATED] = g_signal_new ("tag_activated",
	                                       G_TYPE_FROM_CLASS (object_class),
	                                       G_SIGNAL_RUN_LAST,
	                                       G_STRUCT_OFFSET (LeaftagGtkTagListClass, tag_activated),
	                                       NULL, NULL,
	                                       g_cclosure_marshal_VOID__STRING,
	                                       G_TYPE_NONE,
	                                       1, G_TYPE_STRING);
	signals[TAG_DEACTIVATED] = g_signal_new ("tag_deactivated",
	                                         G_TYPE_FROM_CLASS (object_class),
	                                         G_SIGNAL_RUN_LAST,
	                                         G_STRUCT_OFFSET (LeaftagGtkTagListClass, tag_deactivated),
	                                         NULL, NULL,
	                                         g_cclosure_marshal_VOID__STRING,
	                                         G_TYPE_NONE,
	                                         1, G_TYPE_STRING);
}

static void
leaftag_gtk_tag_list_init (LeaftagGtkTagList *list)
{
	list->priv = g_new0 (LeaftagGtkTagListPriv, 1);

	list->priv->buttons         = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
	list->priv->button_widths   = g_hash_table_new (g_str_hash, g_str_equal);
	list->priv->button_handlers = g_hash_table_new (g_direct_hash, g_direct_equal);

	gtk_box_set_spacing (GTK_BOX (list), 6);
}

static void leaftag_gtk_tag_list_finalize (GObject *obj)
{
	LeaftagGtkTagList *list;

	g_return_if_fail (obj != NULL);
	g_return_if_fail (LEAFTAG_GTK_IS_TAG_LIST (obj));

	list = LEAFTAG_GTK_TAG_LIST (obj);
	g_hash_table_destroy (list->priv->buttons);
	g_hash_table_destroy (list->priv->button_widths);
	g_free (list->priv);

	if (G_OBJECT_CLASS (parent_class)->finalize)
		G_OBJECT_CLASS (parent_class)->finalize (obj);
}

static void
leaftag_gtk_tag_list_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
{
	LeaftagGtkTagList *list;

	if (GTK_WIDGET_CLASS (parent_class)->size_allocate)
		GTK_WIDGET_CLASS (parent_class)->size_allocate (widget, allocation);

	list = LEAFTAG_GTK_TAG_LIST (widget);
	if (allocation->width != list->priv->width)
		reflow (list);
}

GtkWidget *
leaftag_gtk_tag_list_new (void)
{
	return GTK_WIDGET (g_object_new (LEAFTAG_GTK_TYPE_TAG_LIST, 0));
}

void
leaftag_gtk_tag_list_add_tag (LeaftagGtkTagList *list, const gchar *tag)
{
	GtkWidget *button;
	gchar     *name;
	guint      handler;

	button = (GtkWidget *) (g_hash_table_lookup (list->priv->buttons, tag));
	if (button)
		return;

	button = gtk_toggle_button_new_with_label (tag);
	gtk_button_set_relief (GTK_BUTTON (button), GTK_RELIEF_NONE);
	g_object_ref (button);

	handler = g_signal_connect (G_OBJECT (button), "toggled", G_CALLBACK (button_toggled), list);
	g_hash_table_insert (list->priv->button_handlers, button, GUINT_TO_POINTER (handler));

	name = g_strdup (tag);
	list->priv->tags = g_slist_insert_sorted (list->priv->tags, name, (GCompareFunc) strcmp);
	g_object_set_data (G_OBJECT (button), "tag-name", name);

	g_hash_table_insert (list->priv->buttons, name, button);
	gtk_widget_show (button);

	if (GTK_WIDGET_VISIBLE (list))
		reflow (list);
}

void
leaftag_gtk_tag_list_remove_tag (LeaftagGtkTagList *list, const gchar *tag)
{
	GtkWidget *button;
	GSList    *item;

	item = g_slist_find_custom (list->priv->tags, tag, (GCompareFunc) strcmp);
	if (item == NULL)
		return;

	/* Remove the tag from the list */
	list->priv->tags = g_slist_remove_link (list->priv->tags, item);
	button = GTK_WIDGET (g_hash_table_lookup (list->priv->buttons, tag));
	g_slist_free (item);

	g_hash_table_remove (list->priv->button_handlers, button);
	g_hash_table_remove (list->priv->buttons, tag);
	g_hash_table_remove (list->priv->button_widths, tag);

	if (GTK_WIDGET_VISIBLE (list))
		reflow (list);
}

void
leaftag_gtk_tag_list_set_active (LeaftagGtkTagList *list, const gchar *tag)
{
	GtkWidget *button;
	guint handler;

	button = GTK_WIDGET (g_hash_table_lookup (list->priv->buttons, tag));

	if (button == NULL)
		return;

	handler = GPOINTER_TO_UINT (g_hash_table_lookup (list->priv->button_handlers, button));
	g_signal_handler_block (G_OBJECT (button), handler);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), TRUE);
	g_signal_handler_unblock (G_OBJECT (button), handler);
}

void
leaftag_gtk_tag_list_set_inactive (LeaftagGtkTagList *list, const gchar *tag)
{
	GtkWidget *button;
	guint handler;

	button = GTK_WIDGET (g_hash_table_lookup (list->priv->buttons, tag));

	if (button == NULL)
		return;

	handler = GPOINTER_TO_UINT (g_hash_table_lookup (list->priv->button_handlers, button));
	g_signal_handler_block (G_OBJECT (button), handler);
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (button), FALSE);
	g_signal_handler_unblock (G_OBJECT (button), handler);
}

void
leaftag_gtk_tag_list_set_inconsistent (LeaftagGtkTagList *list, const gchar *tag)
{
	GtkWidget *button;

	button = GTK_WIDGET (g_hash_table_lookup (list->priv->buttons, tag));

	if (button == NULL)
		return;

	gtk_toggle_button_set_inconsistent (GTK_TOGGLE_BUTTON (button), TRUE);
}

static void
reflow (LeaftagGtkTagList *list)
{
	gint       width;
	GSList    *i;
	GtkWidget *current_row;
	gint       current_row_items;
	gint       current_row_width;

	if ((GTK_WIDGET_REALIZED (list) == FALSE) || list->priv->frozen)
		return;

	width = GTK_WIDGET (list)->allocation.width;

	/* Deparent all of the buttons */
	for (i = list->priv->tags; i; i = g_slist_next (i)) {
		GtkWidget *button, *parent;
		button = GTK_WIDGET (g_hash_table_lookup (list->priv->buttons, (gchar *) (i->data)));
		parent = gtk_widget_get_parent (button);
		if (parent)
			gtk_container_remove (GTK_CONTAINER (parent), button);
	}

	/* Deparent all of the rows */
	list->priv->rows = g_slist_reverse (list->priv->rows);
	for (i = list->priv->rows; i; i = g_slist_next (i)) {
		GtkWidget *row = GTK_WIDGET (i->data);
		gtk_container_remove (GTK_CONTAINER (list), row);
	}

	/* Finally, free the rows */
	g_slist_free (list->priv->rows);
	list->priv->rows = NULL;

	/* Run through all the tags, packing them into boxes */
	current_row = gtk_hbox_new (FALSE, 6);
	list->priv->rows = g_slist_prepend (list->priv->rows, current_row);
	current_row_items = 0;
	current_row_width = 0;
	for (i = list->priv->tags; i; i = g_slist_next (i)) {
		gchar     *name;
		GtkWidget *button;
		gint       button_width;

		name         = (gchar *) (i->data);
		button       = g_hash_table_lookup (list->priv->buttons, name);
		button_width = GPOINTER_TO_INT (g_hash_table_lookup (list->priv->button_widths, name));
		if (button_width == 0) {
			GtkRequisition size_request;
			gtk_widget_get_child_requisition(button, &size_request);
			button_width = size_request.width;
			g_hash_table_insert (list->priv->button_widths, name, GINT_TO_POINTER (size_request.width));
		}

		if (current_row_items == 0) {
			gtk_box_pack_start (GTK_BOX (current_row), button, FALSE, TRUE, 0);
			current_row_width = button_width;
			current_row_items = 1;
		} else {
			if ((current_row_width + button_width + 6) < width) {
				gtk_box_pack_start (GTK_BOX (current_row), button, FALSE, TRUE, 0);
				current_row_width += (button_width + 6);
				current_row_items += 1;
			} else {
				current_row = gtk_hbox_new (FALSE, 6);
				current_row_items = 1;
				list->priv->rows = g_slist_prepend (list->priv->rows, current_row);
				gtk_box_pack_start (GTK_BOX (current_row), button, FALSE, TRUE, 0);
				current_row_width = button_width;
			}
		}
	}
	list->priv->rows = g_slist_reverse (list->priv->rows);

	/* Pack new rows into the vbox, show children */
	for (i = list->priv->rows; i; i = g_slist_next (i)) {
		GtkWidget *row;

		row = GTK_WIDGET (i->data);
		gtk_widget_show (row);
		gtk_box_pack_start (GTK_BOX (list), row, TRUE, TRUE, 0);
	}

	list->priv->width = width;
}

static void
button_toggled (GtkToggleButton *button, LeaftagGtkTagList *list)
{
	gboolean  toggled;
	gchar    *name;

	toggled = gtk_toggle_button_get_active (button);

	name = g_object_get_data (G_OBJECT (button), "tag-name");

	if (toggled)
		g_signal_emit (G_OBJECT (list), signals[TAG_ACTIVATED], 0, name);
	else
		g_signal_emit (G_OBJECT (list), signals[TAG_DEACTIVATED], 0, name);
}

void
leaftag_gtk_tag_list_freeze (LeaftagGtkTagList *list)
{
	list->priv->frozen = TRUE;
}

void
leaftag_gtk_tag_list_thaw (LeaftagGtkTagList *list)
{
	list->priv->frozen = FALSE;
	reflow (list);
}
